#!/usr/bin/python3

"""
- migrate ODS DB 1.4 to 2.0 (SQLite only)
- migrate ODS configuration
"""
from pathlib import Path
import os
import pwd
import shutil
import sqlite3
import subprocess
import sys

from lxml import etree as ET


ODS_SYSCONFIG_DIR = "/etc/opendnssec"
ODS_CONFIG = os.path.join(ODS_SYSCONFIG_DIR, "conf.xml")
ODS_HOME = "/var/lib/opendnssec"
ODS_ENFORCER_MIGRATED = os.path.join(ODS_HOME, "enforcer.db.migrated")
ODS_BACKUP_DIR = os.path.join(ODS_HOME, "backup")
ODS_SQLITE_MIGRATOR = (
    "/usr/share/opendnssec/enforcer/utils/1.4-2.0_db_convert/convert_sqlite"
)
ALT_MIGRATION_GUIDE = (
    "/usr/share/opendnssec/enforcer/utils/1.4-2.0_db_convert/README.md"
)

OK_CODE = 0
ERROR_CODE = 1

ERROR_MSG = (
    "***ERROR*** Automatic ODS migration failed. Please, follow "
    f"'{ALT_MIGRATION_GUIDE}'. "
    f"Backups for ODS 1.4 can be found in: '{ODS_BACKUP_DIR}'"
)

"""
Typical ODS configuration looks like

<?xml version="1.0" encoding="UTF-8"?>

<Configuration>
        <Enforcer>
                <Privileges>
                        <User>_opendnssec</User>
                        <Group>_opendnssec</Group>
                </Privileges>

                <Datastore><SQLite>/path/kasp.db</SQLite></Datastore>
        </Enforcer>
        ...
"""


def is_systemd_booted():
    # https://www.freedesktop.org/software/systemd/man/sd_booted.html
    systemd_rundir = "/run/systemd/system"
    return not os.path.islink(systemd_rundir) and os.path.isdir(systemd_rundir)


def change_softhsm_mod_name(ods_config):
    # check module name
    # softhsm changed link for opendnssec: libsofthsm.so => libsofthsm2.so
    module = ods_config.find(
        "./RepositoryList/Repository[@name='SoftHSM']/Module"
    )
    if module is None:
        # nothing to do
        return False

    if module.text == "/usr/lib64/softhsm/libsofthsm.so":
        print("Changing SoftHSM module name")
        module.text = "/usr/lib64/softhsm/libsofthsm2.so"
        return True

    return False


def remove_interval(ods_config):
    # remove no longer used Interval
    enforcer = ods_config.find("./Enforcer")
    if enforcer is None:
        return False

    interval = enforcer.find("./Interval")
    if interval is None:
        return False

    print("Removing no longer used Enforcer's Interval")
    enforcer.remove(interval)
    return True


def migrate_config(config_tree):
    ods_config = config_tree.getroot()

    if any((change_softhsm_mod_name(ods_config), remove_interval(ods_config))):
        conf_backup_path = os.path.join(
            ODS_BACKUP_DIR, "conf.xml.1.4.rpmbackup"
        )
        if not Path(conf_backup_path).exists():
            shutil.copy2(ODS_CONFIG, conf_backup_path)
        config_tree.write(ODS_CONFIG, encoding="UTF-8", xml_declaration=True)
        print("ODS config has been migrated")
        return

    print("ODS config is up to date")


def datastore_supported(datastore):
    return datastore.tag in {"SQLite"}


def is_old_sqlitedb(db_path):
    with sqlite3.connect(db_path) as con:
        cursor = con.cursor()
        try:
            cursor.execute(
                "SELECT name FROM sqlite_master WHERE type='table' AND "
                "name='dbadmin';"
            )
        except sqlite3.Error as e:
            raise RuntimeError(e)

        table_name = cursor.fetchone()
        return bool(table_name)


def is_old_datastore(datastore):
    db_type = datastore.tag
    if not datastore_supported(datastore):
        raise ValueError(f"Unsupported DataStore:'{db_type}'")

    if db_type == "SQLite":
        db_path = datastore.text
        print(f"Checking store: '{db_type}':'{db_path}'")
        return is_old_sqlitedb(db_path)

    return False


def migrate_datastore(datastore):
    db_type = datastore.tag
    if not datastore_supported(datastore):
        raise ValueError(f"Unsupported DataStore:'{db_type}'")

    if db_type == "SQLite":
        db_path = datastore.text
        return subprocess.run(
            [
                ODS_SQLITE_MIGRATOR,
                "-i",
                db_path,
                "-o",
                f"{db_path}.migrated",
            ],
            check=False,
        ).returncode

    return ERROR_CODE


def has_backup_db(datastore):
    db_type = datastore.tag
    if not datastore_supported(datastore):
        raise ValueError(f"Unsupported DataStore:'{db_type}'")

    if db_type == "SQLite":
        return Path(
            os.path.join(ODS_BACKUP_DIR, "kasp.db.1.4.rpmbackup")
        ).exists()

    return False


def backup_db(datastore):
    db_type = datastore.tag
    if not datastore_supported(datastore):
        raise ValueError(f"Unsupported DataStore:'{db_type}'")

    if db_type == "SQLite":
        shutil.copy2(
            datastore.text,
            os.path.join(ODS_BACKUP_DIR, "kasp.db.1.4.rpmbackup"),
        )


def replace_db(datastore):
    db_type = datastore.tag
    if not datastore_supported(datastore):
        raise ValueError(f"Unsupported DataStore:'{db_type}'")

    if db_type == "SQLite":
        db_path = datastore.text
        shutil.copy2(f"{datastore.text}.migrated", db_path)


def is_systemd_unit_enabled(unit):
    return (
        subprocess.run(
            ["systemctl", "is-enabled", unit], capture_output=True, check=False
        ).returncode
        == 0
    )


def stop_systemd_units():
    units_to_stop = []
    for unit in ["ods-signerd.service", "ods-enforcerd.service"]:
        if is_systemd_unit_enabled(unit):
            units_to_stop.extend([unit])

    if is_systemd_unit_enabled("ipa.service"):
        units_to_stop.extend(
            ["ipa-ods-exporter.socket", "ipa-ods-exporter.service"]
        )

    for unit in units_to_stop:
        subprocess.run(
            ["systemctl", "stop", unit], capture_output=True, check=False
        )


def start_systemd_units():
    units_to_start = []
    for unit in ["ods-enforcerd.service", "ods-signerd.service"]:
        if is_systemd_unit_enabled(unit):
            units_to_start.extend([unit])

    if is_systemd_unit_enabled("ipa.service"):
        units_to_start.extend(["ipa-ods-exporter.socket"])

    for unit in units_to_start:
        subprocess.run(
            ["systemctl", "start", unit], capture_output=True, check=False
        )


def set_automigrated():
    Path(ODS_ENFORCER_MIGRATED).touch()


def copy_zones():
    zones_path_dst = os.path.join(ODS_HOME, "enforcer", "zones.xml")
    zones_path_src = os.path.join(ODS_SYSCONFIG_DIR, "zonelist.xml")
    print(
        "Copying zones file: '{}' => '{}'".format(
            zones_path_src, zones_path_dst
        )
    )
    shutil.copy2(zones_path_src, zones_path_dst)
    opendnssec_user = pwd.getpwnam("_opendnssec")
    os.chown(zones_path_dst, 0, opendnssec_user.pw_gid)
    os.chmod(zones_path_dst, 0o0660)


def main():
    print(f"Checking for automigration state")
    if Path(ODS_ENFORCER_MIGRATED).exists():
        print("Already automigrated")
        return OK_CODE

    if not is_systemd_booted():
        print(ERROR_MSG)
        return ERROR_CODE

    print(f"Starting automigration")
    parser = ET.XMLParser(remove_comments=False)
    try:
        config_tree = ET.parse(ODS_CONFIG, parser=parser)
    except (FileNotFoundError, ET.ParseError) as e:
        print(f"ERROR: cannot read '{ODS_CONFIG}': {e}")
        return ERROR_CODE

    ods_config = config_tree.getroot()

    for datastore in {"SQLite", "MySQL"}:
        enforcer_datastore = ods_config.find(
            f"./Enforcer/Datastore/{datastore}"
        )
        if enforcer_datastore is not None:
            break

    if enforcer_datastore is None:
        print(ERROR_MSG)
        return ERROR_CODE

    if not datastore_supported(enforcer_datastore):
        print(ERROR_MSG)
        return ERROR_CODE

    try:
        old_datastore = is_old_datastore(enforcer_datastore)
    except RuntimeError as e:
        print(
            f"Unable to detect old datastore version, skipping migration: {e}"
        )
        print(ERROR_MSG)
        return ERROR_CODE

    if not old_datastore:
        print("Nothing to migrate")
        set_automigrated()
        return OK_CODE

    stop_systemd_units()
    print(f"Migrating ODS config: '{ODS_CONFIG}'")
    migrate_config(config_tree)

    print(f"Migrating old ODS datastore")
    if not has_backup_db(enforcer_datastore):
        backup_db(enforcer_datastore)

    result = migrate_datastore(enforcer_datastore)
    if result != 0:
        print(ERROR_MSG)
        return result

    replace_db(enforcer_datastore)
    print("Running ods-migrate")
    result = subprocess.run(["ods-migrate"], check=False).returncode
    if result != 0:
        print(ERROR_MSG)
        return result

    copy_zones()
    start_systemd_units()
    set_automigrated()
    print(f"***SUCCESS*** Automatic ODS migration completed successfully")
    return OK_CODE


if __name__ == "__main__":
    sys.exit(main())
